﻿package {
	
	import flash.display.MovieClip;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	
	
	
	/**
	 * ...
	 * 
	 * 2011/04/28
	 * @author necoEngine bamboo-satellites
	 * @version 
	 * @see http://
	 */
	public class Main extends MovieClip {
		
		private const NUM_FIELDS:uint  = 5;                       // フィールドの総数。表示エリア一つ+隣接する4フィールド。
		private const NUM_CELL_X:uint  = 5;                       // 表示エリアに横に並べるセルの数。
		private const NUM_CELL_Y:uint  = 5;                       // 表示エリアに縦に並べるセルの数。
		private const NUM_CELLS:uint   = NUM_CELL_X * NUM_CELL_Y; // 表示エリアに並ぶセルの総数。
		private const CELL_PIX:uint    = 40;                      // セルの大きさ。正方形。
		private const NUM_PIECES:uint  = 4;                       // ピースの絵柄の数。
		
		private const SLIDE_ALLOW:uint = 5;                       // 「ピースが動かされた」と認識するマウスの最小移動量。
		
		private const TOP:uint    = 0;                // 上方向を表すコード
		private const RIGHT:uint  = 1;                // 右方向を表すコード
		private const BOTTOM:uint = 2;                // 下方向を表すコード
		private const LEFT:uint   = 3;                // 左方向を表すコード
		private const WALL:uint   = 999;              // 番兵。壁をあらわすコード。
		
		private const PIECE1:uint = 1;                // ピースの絵柄を表すコード
		private const PIECE2:uint = 2;                // ピースの絵柄を表すコード
		private const PIECE3:uint = 3;                // ピースの絵柄を表すコード
		private const PIECE4:uint = 4;                // ピースの絵柄を表すコード
		
		private const RESET:uint        = 0;          // フラグリセット
		private const PIECE_PRESS:uint  = 0x00000001; // マウスプレスフラグ
		private const PIECE_SLIDE:uint  = 0x00000002; // スライド検出フラグ
		private const AIM_VERTICAL:uint = 0x00000004; // 縦移動フラグ
		private const AIM_TOP:uint      = 0x00000008; // 上移動フラグ
		private const AIM_LEFT:uint     = 0x00000010; // 左移動フラグ
		
		
		private var adjacencyTable:Vector.<Vector.<int>>;  // 表示エリアのセル毎の隣接情報
		private var piecesFieldData:Vector.<int>;          // 表示エリアのセルに配置するピースの情報
		private var scrollPiecesData:Vector.<int>;         // スクロールさせるピースのグループ
		private var pieces:Vector.<MovieClip>;             // 全てのピースインスタンスの参照を保持
		private var field:Sprite;                          // ピースインスタンスを配置する表示コンテナ。
		private var fieldBounds:Rectangle = new Rectangle( // フィールドの表示領域（マスクのサイズと座標）
			20, 20,                                          // x, y
			NUM_CELL_X * CELL_PIX, NUM_CELL_Y * CELL_PIX     // widht, height
		);
		
		private var ctrlStatus:uint;                       // ユーザの操作状況を表すビットマスクフラグ
		private var mouseLocation:Point;                   // マウス座標を保存
		private var preMouseLocation:Point;                // mouseLocationよりも1フレーム前の座標を保存
		private var scrollValidBounds:Rectangle;           // スクロール処理を継続可能なマウスの有効領域
		
		
		
		public function Main():void {
			
			if (stage) init();
			else addEventListener(Event.ADDED_TO_STAGE, init);
			
		} // constructor close
		
		
		private function init():void {
			
			removeEventListener(Event.ADDED_TO_STAGE, init);
			
			// フィールドのセル同士の隣接情報テーブルを生成する
			makeAdjacencyTable();
			
			// フィールドに配置するピースの絵柄データを設定
			fieldDataInit();
			
			// ピースを配置する表示コンテナ
			field = new Sprite();
			addChild(field);
			field.x = fieldBounds.x;
			field.y = fieldBounds.y;
			// フィールドにマスクを生成、適用。
			field.mask = getMask(fieldBounds);
			
			// フィールドにピースインスタンスを配置。配置したインスタンスの参照を配列化。
			piecesInit();
			
			// 配置したピースインスタンスの絵柄を更新
			piecesUpdate(piecesFieldData);
			
			// マウスプレスイベント設定
			ctrlStatus = RESET;
			field.addEventListener(MouseEvent.MOUSE_DOWN, onPiecePress);
			
		} // init() close
		
		
		
//================================================================================
// データ操作
//================================================================================
		private function makeAdjacencyTable():void {
			/**
			 * 機能：フィールドにある各セルの隣接リストを作成する。
			 * 引数：なし
			 * 戻り値：なし
			 * 備考：二次元目の要素のインデックスは、対象セルを基準にした方向をあらわしている。
			 *       [0]上 [1]右 [2]下 [3]左（上からの時計回り）
			 */
			
			var countX:uint, countY:uint = 0;
			adjacencyTable = new Vector.<Vector.<int>>(NUM_CELLS, true);
			
			for (var i:uint = 0; i < NUM_CELLS; i++) {
				adjacencyTable[i] = new Vector.<int>(4, true);
				
				// countXが横幅一杯に到達したかどうかをチェック
				if (countX >= NUM_CELL_X) {
					countX = 0;
					countY++;
				}
				
				adjacencyTable[i][TOP]    = (countY > 0)?              i - NUM_CELL_X : WALL;
				adjacencyTable[i][RIGHT]  = (countX < NUM_CELL_X - 1)? i + 1          : WALL;
				adjacencyTable[i][BOTTOM] = (countY < NUM_CELL_Y - 1)? i + NUM_CELL_Y : WALL;
				adjacencyTable[i][LEFT]   = (countX > 0)?              i - 1          : WALL;
				
				countX++;
			}
			
		} // makeAdjacencyTable() close
		
		
		private function fieldDataInit():void {
			/**
			 * 機能：フィールドに並べるピース絵柄をランダムで決定して
			 *       piecesFieldData配列に絵柄コードをセットする。
			 * 引数：なし
			 * 戻り値：なし
			 * 備考：
			 */
			
			piecesFieldData = new Vector.<int>(NUM_CELLS, true);
			for (var i:uint = 0; i < NUM_CELLS; i++) piecesFieldData[i] = Math.random() * NUM_PIECES + 1;
			
		} // fieldDataUpdate() close
		
		
		private function slideGroupSearch(target:int, vertical:Boolean):Vector.<int> {
		/**
		 * 機能：スクロールさせるピースのコードをグループ化して配列で返す。
		 * 引数：第1引数 プレスされているピースのインデックス
		 *       第2引数 縦スライドさせたかの真偽値。縦の場合真。
		 * 戻り値：ピースのコードをグループ化したVector配列
		 * 備考：動かすだけならホントは順番は気にしなくてよい。
		 */
		
		var i:int, start:int, end:int, index:int = target;
		var group:Vector.<int> = new Vector.<int>();
		
		if (vertical) {
			//縦にスライドする場合
			
			// ターゲットがある列で、表示エリア一番上にあたるセルのインデックスを探索
			while (adjacencyTable[index][TOP] != WALL) index = adjacencyTable[index][TOP];
			
			// 表示エリア上側スクロールエリアのインデックスを抽出
			start = NUM_CELLS * (TOP + 1) + index;
			end   = start + NUM_CELLS - NUM_CELL_X;
			for (i = start; i <= end; i += NUM_CELL_X) group.push(i);
			
			// 表示エリアのインデックスを抽出
			end = NUM_CELLS - NUM_CELL_X + index;
			for (i = index; i <= end; i += NUM_CELL_X) group.push(i);
			
			// 表示エリア下側スクロールエリアのインデックスを抽出
			start = NUM_CELLS * (BOTTOM + 1) + index;
			end   = start + NUM_CELLS - NUM_CELL_X;
			for (i = start; i <= end; i += NUM_CELL_X) group.push(i);
			
		} else {
			//横にスライドする場合
			
			// ターゲットがある行で、表示エリア一番左にあたるセルのインデックスを探索
			while (adjacencyTable[index][LEFT] != WALL) index = adjacencyTable[index][LEFT];
			
			// 表示エリア左側スクロールエリアのインデックスを抽出
			start = NUM_CELLS * (LEFT + 1) + index;
			end   = start + NUM_CELL_X;
			for (i = start; i < end; i++) group.push(i);
			
			// 表示エリアのインデックスを抽出
			end = NUM_CELL_X + index;
			for (i = index; i < end; i++) group.push(i);
			
			// 表示エリア右側スクロールエリアのインデックスを抽出
			start = NUM_CELLS * (RIGHT + 1) + index;
			end   = start + NUM_CELL_X;
			for (i = start; i < end; i++) group.push(i);
			
		}
		
		return group;
		
	} // slideGroupSearch() close
		
		
		
//================================================================================
// 描画処理
//================================================================================
		private function piecesInit():void {
			/**
			 * 機能：ピースインスタンスを生成してフィールドに配置。
			 *       配置したインスタンスの参照はVector配列にまとめる。
			 * 引数：なし
			 * 戻り値：なし
			 * 備考：
			 */
			
			var index:int, x:int, y:int, xAim:int = 0, yAim:int = 0;
			var fieldWidth:int  = NUM_CELL_X * CELL_PIX;
			var fieldHeight:int = NUM_CELL_Y * CELL_PIX;
			
			pieces = new Vector.<MovieClip>(NUM_CELLS * NUM_FIELDS, true);
			
			for (var i:int = 0; i < NUM_FIELDS; i++) {
				
				index = 0;
				x = 0;
				y = 0;
				
				for (var j:int = 0; j < NUM_CELLS; j++) {
					
					if (x >= NUM_CELL_X) {
						x = 0;
						y++;
					}
					
					// インデックスを算出
					index = i * NUM_CELLS + j;
					
					pieces[index] = new PieceMC();
					field.addChild(pieces[index]);
					
					pieces[index].x = x * CELL_PIX + xAim;
					pieces[index].y = y * CELL_PIX + yAim;
					
					x++;
				}
				
				// 該当フィールドによって、ピースを配置する基準座標を変える。
				switch (i) {
					case TOP   : xAim = 0;           yAim = -fieldHeight; break;
					case RIGHT : xAim = fieldWidth;  yAim = 0;            break;
					case BOTTOM: xAim = 0;           yAim = fieldHeight;  break;
					case LEFT  : xAim = -fieldWidth; yAim = 0;            break;
				}
			}
			
		} // piecesInit() close
		
		
		private function piecesUpdate(data:Vector.<int>):void {
			/**
			 * 機能：ピースの絵柄をフィールド情報にしたがって変更する
			 * 引数：ピース絵柄を更新する為のフィールドデータ配列。
			 * 戻り値：なし
			 * 備考：
			 */
			
			data.forEach(
				function (element:int, index:int, array:Vector.<int>):void {
					pieces[index].gotoAndStop(element);
					pieces[index + (TOP    + 1) * NUM_CELLS].gotoAndStop(element);
					pieces[index + (RIGHT  + 1) * NUM_CELLS].gotoAndStop(element);
					pieces[index + (BOTTOM + 1) * NUM_CELLS].gotoAndStop(element);
					pieces[index + (LEFT   + 1) * NUM_CELLS].gotoAndStop(element);
				});
			
		} // piecesUpdate() close
		
		
		private function getMask(bounds:Rectangle):Sprite {
			/**
			 * 機能：指定された位置とサイズでマスク用の短形を描画して返す。
			 * 引数：マスクの描画指示。
			 * 戻り値：マスク用に描画した短形
			 * 備考：
			 */
			
			var mask:Sprite = new Sprite();
			mask.graphics.beginFill(0xFF0000);
			mask.graphics.drawRect(bounds.x, bounds.y, bounds.width, bounds.height);
			mask.graphics.endFill();
			return mask;
			
		} // getMask() close
		
		
		private function onPiecesScroll(event:Event):void {
			/**
			 * 機能：マウスの移動量に合わせてピースを移動する。
			 * 引数：イベントオブジェクト
			 * 戻り値：
			 * 備考：
			 */
			
			var i:int, len:uint, index:int, dist:Number;
			
			// マウス座標が有効領域外へ出たかどうかを評価。出たら移動処理を終了する
			if (mouseX <= scrollValidBounds.left || mouseX >= scrollValidBounds.right ||
				  mouseY <= scrollValidBounds.top  || mouseY >= scrollValidBounds.bottom) {
				onPieceRelease();
			}
			
			//移動処理
			if (ctrlStatus & AIM_VERTICAL) {
				//縦移動処理
				
				dist = field.mouseY - preMouseLocation.y; // マウスが移動した距離を計測
				preMouseLocation.y = field.mouseY;        // 現在の座標を保存
				
				// スクロールするピースグループの座標に移動距離を反映
				len = scrollPiecesData.length;
				for (i = 0; i < len; i++) {
					index = scrollPiecesData[i];
					pieces[index].y += dist;
				}
				
			} else {
				//横移動処理
				
				dist = field.mouseX - preMouseLocation.x;
				preMouseLocation.x = field.mouseX;
				
				len = scrollPiecesData.length;
				for (i = 0; i < len; i++) {
					index = scrollPiecesData[i];
					pieces[index].x += dist;
				}
			}
			
			// マウスがリリースされたらループ終了
			if (!(ctrlStatus & PIECE_PRESS)) {
				field.removeEventListener(Event.ENTER_FRAME, onPiecesScroll);
				piecesPutBack();
			}
			
		} // onPiecesScroll() close
		
		
		private function piecesPutBack():void {
			/**
			 * 機能：移動させたピースグループを元の位置に戻す。
			 * 引数：マウスの移動量
			 * 戻り値：
			 * 備考：
			 */
			
			var i:int, index:int, cell:uint;
			var len:uint  = scrollPiecesData.length;
			var px:Number = -NUM_CELL_X * CELL_PIX; // 配置開始x座標
			var py:Number = -NUM_CELL_Y * CELL_PIX; // 配置開始y座標
			
			//ピースを元の位置に戻す。
			for (i = 0; i < len; i++) {
				index = scrollPiecesData[i];
				
				if (ctrlStatus & AIM_VERTICAL) {
					//縦移動の場合
					pieces[index].y = py;
					py += CELL_PIX;
				} else {
					//横移動の場合
					pieces[index].x = px;
					px += CELL_PIX;
				}
			}
			
			field.addEventListener(MouseEvent.MOUSE_DOWN, onPiecePress);
			
		} // piecesPutBack() close
		
		
		
//================================================================================
// ユーザ操作処理
//================================================================================
		private function onPiecePress(event:MouseEvent):void {
			/**
			 * 機能：マウスリリースのイベントと、スライド操作を監視するためのイベントを設定する。
			 * 引数：イベントオブジェクト
			 * 戻り値：
			 * 備考：名前はonPiecePressだけど、fieldがターゲット
			 */
			
			ctrlStatus |= PIECE_PRESS; // マウスプレスフラグ真
			field.removeEventListener(MouseEvent.MOUSE_DOWN, onPiecePress);
			
			// プレス時のマウス座標を保存する
			mouseLocation = new Point(field.mouseX, field.mouseY);
			
			// マウスリリースイベント
			field.addEventListener(MouseEvent.MOUSE_UP, onPieceRelease);
			
			// スライド操作を監視するためのイベント
			ctrlStatus &= ~ PIECE_SLIDE;
			if (ctrlStatus & PIECE_PRESS) field.addEventListener(MouseEvent.MOUSE_MOVE, onPieceSlideKeeper);
			
		} // onPiecePress() close
		
		
		private function onPieceRelease(event:MouseEvent = null):void {
			/**
			 * 機能：リスナー関数の削除と、フラグの初期化を行う。
			 * 引数：イベントオブジェクト
			 * 戻り値：
			 * 備考：
			 */
			
			field.removeEventListener(MouseEvent.MOUSE_UP, onPieceRelease);
			
			if (ctrlStatus & PIECE_SLIDE) {
				// スライド操作されていた場合
				ctrlStatus &= ~PIECE_PRESS;
				ctrlStatus &= ~PIECE_SLIDE;
			} else {
				// スライド操作されていなかった場合（すぐにリリースされた場合など）
				field.removeEventListener(MouseEvent.MOUSE_MOVE, onPieceSlideKeeper);
				
				ctrlStatus = RESET;
			  field.addEventListener(MouseEvent.MOUSE_DOWN, onPiecePress);
			}
			
		} // onPieceRelease() close
		
		
		private function onPieceSlideKeeper(event:MouseEvent):void {
			/**
			 * 機能：プレス後のマウスの動きを監視し、マウスの動いた量と方向を判定してスクロール処理へ移行する。
			 * 引数：イベントオブジェクト
			 * 戻り値：
			 * 備考：
			 */
			
			var index:int = pieces.indexOf(event.target);
			
			// スライド操作を検出しているかどうかを評価
			if (!(ctrlStatus & PIECE_SLIDE)) {
				
				//縦スクロール発生判定
				if (field.mouseY > mouseLocation.y + SLIDE_ALLOW || field.mouseY < mouseLocation.y - SLIDE_ALLOW) {
					ctrlStatus |= PIECE_SLIDE;  // スライド操作発生フラグ真
					ctrlStatus |= AIM_VERTICAL; // 縦スクロールフラグ真
					
					(field.mouseY < mouseLocation.y - SLIDE_ALLOW)? ctrlStatus |= AIM_TOP: ctrlStatus &= ~AIM_TOP; // 上下の判定
					
					piecesScrollInit(index);
				}
				
				//横スクロール発生判定
				if (field.mouseX > mouseLocation.x + SLIDE_ALLOW || field.mouseX < mouseLocation.x - SLIDE_ALLOW) {
					ctrlStatus |= PIECE_SLIDE;   // スライド操作発生フラグ真
					ctrlStatus &= ~AIM_VERTICAL; // 縦スクロールフラグ偽
					
					(field.mouseX < mouseLocation.x - SLIDE_ALLOW)? ctrlStatus |= AIM_LEFT: ctrlStatus &= ~AIM_LEFT; // 左右の判定
					
					piecesScrollInit(index);
				}
			}
			
		} // onPieceSlideKeeper() close
		
		
		private function piecesScrollInit(index:int):void {
			/**
			 * 機能：スクロール処理のための初期化処理
			 * 引数：プレスされているピースのインデックス
			 * 戻り値：なし
			 * 備考：
			 */
			
			field.removeEventListener(MouseEvent.MOUSE_MOVE, onPieceSlideKeeper);
			
			// 直前の座標情報を保存
			preMouseLocation = new Point(mouseLocation.x, mouseLocation.y);
			
			// スクロール処理を継続可能なマウスポインタの有効領域を設定。
			scrollValidBounds = pieces[index].getBounds(stage);
			if (ctrlStatus & AIM_VERTICAL) {
				scrollValidBounds.top    = fieldBounds.top;
				scrollValidBounds.bottom = fieldBounds.bottom;
			} else {
				scrollValidBounds.left  = fieldBounds.left;
				scrollValidBounds.right = fieldBounds.right;
			}
			
			// スクロール処理するピースをグループ化
			scrollPiecesData = slideGroupSearch(index, (ctrlStatus & AIM_VERTICAL)? true: false);
			
			// スクロール処理へ移行
			field.addEventListener(Event.ENTER_FRAME, onPiecesScroll);
			
		} // piecesScrollInit() close
	}
}